/**
 * $RCSfile: InstantFeedsPlugin.java,v $
 * $Revision: 3117 $
 * $Date: 2005-11-25 22:57:29 -0300 (Fri, 25 Nov 2005) $
 *
 * Copyright (C) 2004 Jive Software. All rights reserved.
 *
 * This software is published under the terms of the GNU Public License (GPL),
 * a copy of which is included in this distribution.
 */

package org.jivesoftware.plugins.instantFeeds;

import org.jivesoftware.openfire.container.Plugin;
import org.jivesoftware.openfire.container.PluginManager;
import org.jivesoftware.openfire.XMPPServer;
import org.jivesoftware.openfire.SessionManager;
import org.jivesoftware.openfire.group.Group;
import org.jivesoftware.openfire.group.GroupManager;
import org.jivesoftware.openfire.group.GroupNotFoundException;
import org.jivesoftware.openfire.user.UserNotFoundException;
import org.jivesoftware.openfire.user.UserAlreadyExistsException;
import org.jivesoftware.openfire.auth.UnauthorizedException;
import org.jivesoftware.openfire.auth.AuthToken;
import org.jivesoftware.openfire.auth.AuthFactory;
import org.jivesoftware.openfire.net.VirtualConnection;
import org.jivesoftware.openfire.session.Session;
import org.jivesoftware.openfire.session.LocalClientSession;
import org.jivesoftware.openfire.interceptor.PacketInterceptor;
import org.jivesoftware.openfire.interceptor.PacketRejectedException;
import org.jivesoftware.openfire.interceptor.InterceptorManager;
import org.jivesoftware.util.JiveGlobals;
import org.jivesoftware.util.Log;
import org.jivesoftware.util.JiveConstants;
import org.jivesoftware.util.StringUtils;
import org.jivesoftware.database.DbConnectionManager;
import org.jivesoftware.database.SequenceManager;
import org.jivesoftware.admin.AuthCheckFilter;

import org.xmpp.packet.*;
import org.xmpp.component.Component;
import org.xmpp.component.ComponentManager;
import org.xmpp.component.ComponentException;
import org.xmpp.component.ComponentManagerFactory;
import org.htmlparser.Parser;
import org.htmlparser.NodeFilter;
import org.htmlparser.Tag;
import org.htmlparser.util.*;
import org.htmlparser.filters.HasAttributeFilter;
import org.htmlparser.filters.TagNameFilter;
import org.htmlparser.filters.AndFilter;

import java.io.File;
import java.io.IOException;
import java.util.concurrent.ConcurrentLinkedQueue;
import java.util.concurrent.ConcurrentMap;
import java.util.concurrent.ConcurrentHashMap;
import java.util.*;
import java.util.Date;
import java.sql.*;
import java.net.URL;
import java.net.MalformedURLException;
import java.net.InetAddress;
import java.net.UnknownHostException;

import com.sun.syndication.fetcher.impl.FeedFetcherCache;
import com.sun.syndication.fetcher.impl.HashMapFeedInfoCache;
import com.sun.syndication.fetcher.FeedFetcher;
import com.sun.syndication.fetcher.FetcherException;
import com.sun.syndication.feed.synd.SyndFeed;
import com.sun.syndication.feed.synd.SyndEntry;
import com.sun.syndication.feed.synd.SyndContent;
import com.sun.syndication.io.FeedException;

/**
 * A plugin for Openfire that:
 * <ul>
 *  <li>
 *      proactively polls RSS / Atom feeds at regular intervals looking for for updates.
 *  </li>
 *
 *  <li>
 *      enables users to subscribe to RSS / Atom feeds via a bot interface and receive
 *      notifications of RSS and Atom feed updates. The plugin can be accessed by sending messages
 *      to 'feeds.$yourserver$' or 'feeds@$yourserver$'.
 *  </li>
 *
 *  <li>
 *      enables publishers to send standard XML-RPC 'pings' when content is updated. The XML-RPC endpoint is:
 *      http://$yourserver:9090/plugins/instantfeeds/xmlrpc/.
 *  </li>
 *
 * </ul>
 *
 * The following properties can be customized:
 * <ul>
 *  <li>
 *      plugin.instantFeeds.serviceName: the name of the service you want users to interact with,
 *      default is 'feeds'. Note that an account will be automatically created by the plugin
 *      if it doesn't already exist.
 *  </li>
 *  <li>
 *      plugin.instantFeeds.servicePassword: the password for the service / user created by the
 *      system. Default is 'password'.
 *  </li>
 *  <li>
 *      plugin.instantFeeds.allowedUsers: a comma delimited list of users who are allowed to
 *      interact with the service. Defaults to an empty string which means all users can
 *      interact with the system.
 *  </li>
 *  <li>
 *      plugin.instantFeeds.allowedGroups: a comma delimited list of groups who are allowed
 *      to interact with the service. Defaults to an empty string which means all groups
 *      can interact with the system.
 *  
 *  <li>
 *      plugin.instantFeeds.numThreads: the number of threads you want each subservice (notifiers,
 *      subscribers, and pollers) to use, default is 2.
 *  </li>
 *  <li>
 *      plugin.instantFeeds.pollInterval: the number of milliseconds to wait between each poll of
 *      existing feeds. Default is 15 minutes (900000 milliseconds).
 *  </li>
 *  <li>
 *      plugin.instantFeeds.serverName: the path to the HTTP server that will handle verification
 *      clicks. Defaults to http://yourserver.com:9090/plugins/instantfeeds/api. Note that this
 *      server name can be shortened through the use of mod_proxy and mod_rewrite, see the readme
 *      for more information.
 *  </li>
 *  <li>
 *      plugin.instantFeeds.proxyHost: the host name of the proxy server to use when making HTTP
 *      requests for RSS / Atom feeds.  Optional.
 *  </li>
 *  <li>
 *      plugin.instantFeeds.proxyPort: the port of the proxy server to use when making HTTP
 *      requests for RSS / Atom feeds. Optional.
 *  </li>
 * </ul>
 */
public class InstantFeedsPlugin implements Plugin, Component, PacketInterceptor {

    public static final int JIVEIDFEED = 127;
    public static final int JIVEIDFEEDUSER = 128;

    private static final int NUMBER_NOTIFIERS = 2;

    private static final long POLL_INTERVAL = 900000; // fifteen minutes

    private static final String CREATE_USER =
            "INSERT INTO instantFeedUser (userID, jid, hibernating) VALUES (?, ?, 0)";

    private static final String CREATE_FEED =
            "INSERT INTO instantFeedFeed (feedID, feedName, feedURL, mostRecentEntry) VALUES (?,?,?,?)";

    private static final String CREATE_SUBSCRIPTION =
            "INSERT INTO instantFeedUserToFeed " +
                    "(userid, feedID, verified, mostRecentEntry) VALUES(?,?,?,?)";

    private static final String GET_USERS_BY_FEED =
            "SELECT u.userID,u.jid,u.hibernating FROM instantFeedUser u, instantFeedUserToFeed j " +
            "WHERE u.userID = j.userID AND j.feedID = ? AND j.verified = 1";

    private static final String GET_USERS_COUNT_BY_FEED =
            "SELECT count(u.userID) FROM instantFeedUser u, instantFeedUserToFeed j, instantFeedFeed f " +
            "WHERE u.userID = j.userID AND j.feedID = f.feedID AND f.feedURL = ?";

    private static final String GET_FEEDS_BY_USER =
            "SELECT f.feedID, f.feedName, f.feedURL, f.mostRecentEntry " +
            "FROM instantFeedFeed f, instantFeedUserToFeed j " +
            "WHERE f.feedID = j.feedID AND j.userID = ? AND j.verified = 1";

    private static final String GET_FEEDS = "SELECT feedID, feedName, feedURL, mostRecentEntry " +
            "FROM instantFeedFeed";

    private static final String GET_FEED_BY_URL = "SELECT feedID, feedName, feedURL, mostRecentEntry " +
            "FROM instantFeedFeed WHERE feedURL = ?";

    private static final String GET_USER_BY_JID = "SELECT userID, jid, hibernating " +
            "FROM instantFeedUser where jid = ?";

    private static final String DELETE_SUBSCRIPTION =
            "DELETE FROM instantFeedUserToFeed WHERE userID = ? AND feedID = ?";

    private static final String UPDATE_FEED = "UPDATE instantFeedFeed SET feedName = ?, feedURL = ? " +
            "mostRecentEntry = ? WHERE feedID = ?";

    private static final String UPDATE_SUBSCRIPTION_VERIFIED = "UPDATE instantFeedUserToFeed " +
            "SET verified = 1 WHERE userID = ? AND feedID = ?";

    private static final String UPDATE_SUBSCRIPTION_LAST_ENTRY = "UPDATE instantFeedUserToFeed " +
            "SET mostRecentEntry = ? WHERE feedID = ? AND userID = ?";

    private static final String UPDATE_USER_SLEEP_STATE = "UPDATE instantFeedUser SET hibernating = ? " +
            "WHERE userID = ?";

    private static final String GET_SUBSCRIPTION_LAST_ENTRY = "SELECT mostRecentEntry " +
            "FROM instantFeedUserToFeed WHERE userID = ? AND feedID = ?";

    
    private String serviceName; // the name of the component
    private String servicePassword; // the password for the account
    private String serverName; // path to http://yourserver.com/api which forwards to :9090/plugins/instantfeeds/api
    private String proxyHost;
    private int proxyPort;
    private JID bot; // the jid of the bot that sends / receives messages
    private PluginManager pluginManager;
    private GroupManager groupManager;
    private ComponentManager componentManager;

    private List<JID> allowedUsers;
    private List<Group> allowedGroups;

    private Collection<SendNotificationsTask> notifiers;

    private Collection<NewFeedsTask> subscribers;

    private Collection<PollFeedsTask> pollers;

    private LoadFeedsTask loader;

    /**
     * Queue for feeds that have been updated and need to be
     * distributed to users.
     */
    private Queue<Feed> updatedFeeds;

    /**
     * Queue for new feeds that need to be verified before being
     * added to the datastore.
     */
    private Queue<FeedSubscriptionRequest> newFeeds;


    /**
     * Queue for feeds that need to be polled by the pollers.
     */
    private Queue<Feed> existingFeeds;


    /**
     * Map of MD5 hash of feed url + jid to feed subscription requests.
     */
    private ConcurrentMap<String, FeedSubscriptionRequest> nonVerifiedRequests;


    public void initializePlugin(PluginManager manager, File pluginDirectory) {
        pluginManager = manager;
        serviceName = JiveGlobals.getProperty("plugin.instantFeeds.serviceName", "feeds");
        servicePassword = JiveGlobals.getProperty("plugin.instantFeeds.servicePassword", "password");
        proxyHost = JiveGlobals.getProperty("plugin.instantFeeds.proxyHost", "");
        proxyPort = JiveGlobals.getIntProperty("plugin.instantFeeds.proxyPort", -1);
        int numberConnections = JiveGlobals.getIntProperty("plugin.instantFeeds.numThreads", NUMBER_NOTIFIERS);
        long pollingInterval = JiveGlobals.getLongProperty("plugin.instantFeeds.pollInterval", POLL_INTERVAL);

        componentManager = ComponentManagerFactory.getComponentManager();
        try {
            componentManager.addComponent(serviceName, this);
        }
        catch (Exception e) {
            Log.error(e);
        }
        
        groupManager = GroupManager.getInstance();

        AuthCheckFilter.addExclude("/api/xmlrpc");
        AuthCheckFilter.addExclude("/api/http/subscriptions");
        
        serverName = JiveGlobals.getProperty("plugin.instantFeeds.serverName",
                "http://" + componentManager.getServerName() + ":9090/plugins/instantfeeds/api");

        allowedUsers = userStringToList(JiveGlobals.getProperty("plugin.instantFeeds.allowedUsers", ""));

        allowedGroups = groupStringToList(JiveGlobals.getProperty("plugin.instantFeeds.allowedGroups", ""));

        InterceptorManager.getInstance().addInterceptor(this);
        bot = new JID(serviceName, componentManager.getServerName(), "bot");

        try {
            XMPPServer.getInstance().getUserManager().getUser(serviceName);
        } catch (UserNotFoundException e) {
            // attempt to create the user if he / she doesn't exist
            try {
                XMPPServer.getInstance().getUserManager().createUser(serviceName,
                        servicePassword, serviceName, null);
            } catch (UserAlreadyExistsException e1) {
                // ignore...
            } catch (UnsupportedOperationException uoe) {
                // ignore...
            }

        }


        try {
            AuthToken token = AuthFactory.authenticate(serviceName, servicePassword);
            SessionManager sessionManager = SessionManager.getInstance();

            // create a pseudo session so that wildfire thinks this user is online
            LocalClientSession session = sessionManager.createClientSession(
                    new FakeConnection());

            session.setAuthToken(token, "bot");
            session.setPresence(new Presence());

            // tell AIM, MSN and Yahoo that the bot is online
            Presence showAvailableAIM = new Presence();
            showAvailableAIM.setTo(new JID("aim." + componentManager.getServerName()));
            showAvailableAIM.setFrom(bot);
            componentManager.sendPacket(this, showAvailableAIM);

            Presence showAvailableMSN = new Presence();
            showAvailableMSN.setTo(new JID("msn." + componentManager.getServerName()));
            showAvailableMSN.setFrom(bot);
            componentManager.sendPacket(this, showAvailableMSN);

            Presence showAvailableYahoo = new Presence();
            showAvailableYahoo.setTo(new JID("yahoo." + componentManager.getServerName()));
            showAvailableYahoo.setFrom(bot);
            componentManager.sendPacket(this, showAvailableYahoo);



        } catch (Exception e) {
            Log.error("error sending online presence" + e.toString());
            e.printStackTrace();
        }


        nonVerifiedRequests = new ConcurrentHashMap<String, FeedSubscriptionRequest>();

        // NOTIFIERS
        updatedFeeds = new ConcurrentLinkedQueue<Feed>();
        notifiers = new ArrayList<SendNotificationsTask>(numberConnections);
        for (int i = 0; i < numberConnections; i++) {
            SendNotificationsTask notifier = new SendNotificationsTask(i);
            notifiers.add(notifier);
            new Thread(notifier).start();
        }

        // GREETERS
        newFeeds = new ConcurrentLinkedQueue<FeedSubscriptionRequest>();
        subscribers = new ArrayList<NewFeedsTask>(numberConnections);
        for (int i = 0; i < numberConnections; i++) {
            NewFeedsTask subscriber = new NewFeedsTask(i);
            subscribers.add(subscriber);
            new Thread(subscriber).start();
        }


        // LOADER
        existingFeeds = new ConcurrentLinkedQueue<Feed>();
        loader = new LoadFeedsTask(pollingInterval);
        new Thread(loader).start();


        // FETCHERS
        pollers = new ArrayList<PollFeedsTask>(numberConnections);
        for (int i = 0; i < numberConnections; i++) {
            PollFeedsTask poller = new PollFeedsTask(i);
            pollers.add(poller);
            new Thread(poller).start();
        }

        // TODO: create a task that expires old entries in nonVerifiedRequests


    }

    public void destroyPlugin() {
        try {
            componentManager.removeComponent(serviceName);
        }
        catch (Exception e) {
            Log.error(e);
        }
        componentManager = null;
        pluginManager = null;
        allowedUsers.clear();

        if (notifiers != null) {
            for (SendNotificationsTask notifier : notifiers) {
                notifier.stop();
            }
        }

        if (subscribers != null) {
            for (NewFeedsTask subscriber : subscribers) {
                subscriber.stop();
            }
        }

        if (loader != null) {
            loader.stop();
        }

        if (pollers != null) {
            for (PollFeedsTask poller : pollers) {
                poller.stop();
            }
        }

    }


    public String getName() {
        return pluginManager.getName(this);
    }

    public String getDescription() {
        return pluginManager.getDescription(this);
    }

    public void initialize(JID jid, ComponentManager componentManager) throws ComponentException {
        if (componentManager != null) {
            this.componentManager = componentManager;
        }
    }

    public void start() {}

    public void shutdown() {}

    public String getPublicServer() {
        return serverName;
    }


    public void interceptPacket(Packet packet, Session session, boolean incoming, boolean processed) throws PacketRejectedException {
        if (!processed && incoming &&
                packet.getTo() != null &&
                packet.getTo().getNode() != null && 
                packet.getTo().getNode().equalsIgnoreCase(bot.getNode()) &&
                packet.getTo().getDomain().equalsIgnoreCase(bot.getDomain())) {
            
                Log.debug(packet.toXML());
                processPacket(packet);
        }
    }

    public void processPacket(Packet packet) {
        if (packet instanceof Message) {
            // Respond to incoming messages
            Message message = (Message)packet;
            processMessage(message);
        } else if (packet instanceof Presence) {
            // Respond to presence subscription request or presence probe
            Presence presence = (Presence) packet;
            processPresence(presence);
        }

    }

    /**
     * Method that filters messages looking for the following commands:
     * <ul>
     *  <li>list</li>
     *  <li>subscribe $url</li>
     *  <li>unsubscribe $url</li>
     *  <li>help</li>
     * </ul>
     * @param message the message to filter
     */
    private void processMessage(Message message) {

        JID address = new JID(message.getFrom().toBareJID());

        // do single allowed user check
        if (!allowedUsers.isEmpty() && !allowedUsers.contains(address)) {
            Message error = new Message();
            if (message.getID() != null) {
                error.setID(message.getID());
            }
            error.setError(PacketError.Condition.not_allowed);
            error.setTo(message.getFrom());
            error.setFrom(message.getTo());
            error.setBody("Not allowed to interact with instantFeeds.");
            try {
                componentManager.sendPacket(this, error);
            }
            catch (Exception e) {
                Log.error(e);
            }

            return;
        }

        // do group check
        if (!allowedGroups.isEmpty()) {
            boolean foundUser = false;
            for (Group group : allowedGroups) {
                if (group.isUser(address)) {
                    foundUser = true;
                    break;
                }
            }
            if (!foundUser) {
                Message error = new Message();
                if (message.getID() != null) {
                    error.setID(message.getID());
                }
                error.setError(PacketError.Condition.not_allowed);
                error.setTo(message.getFrom());
                error.setFrom(message.getTo());
                error.setBody("Not allowed to interact with instantFeeds.");
                try {
                    componentManager.sendPacket(this, error);
                }
                catch (Exception e) {
                    Log.error(e);
                }

                return;
            }
        }




        if (message.getType() == Message.Type.chat || message.getType() == Message.Type.normal) {
            // if 'list' or 'subscribe' or 'unsubscribe'
            if (message.getBody() != null && message.getBody().startsWith("list")) {

                FeedUser user = loadUser(message.getFrom().toBareJID());
                List<Feed> feeds = getFeeds(user);

                Message reply = new Message();
                reply.setType(Message.Type.chat);
                reply.setTo(message.getFrom());
                reply.setFrom(bot);
                if (feeds.size() > 0) {
                    StringBuffer sb = new StringBuffer("I've got you down for the following feeds:");
                    for (Feed feed : feeds) {
                        sb.append(System.getProperty("line.separator"))
                          .append(feed.getName())
                          .append(" ( ")
                          .append(feed.getURL())
                          .append(" )");
                    }
                    reply.setBody(sb.toString());
                } else {
                    reply.setBody("Dude, you totally haven't signed up for any feeds yet.");
                }
                try {
                    componentManager.sendPacket(this, reply);
                }
                catch (Exception e) {
                    Log.error(e);
                }

            } else if (message.getBody() != null && message.getBody().startsWith("last ")) {
                String url = message.getBody().substring("last ".length()).trim();
                try {
                    Feed feed = loadFeed(url);
                    FeedUser user = loadUser(message.getFrom().toBareJID());
                    sendNotification(user, feed, true);
                } catch (Exception e) {
                    Log.error(e);
                }


            } else if (message.getBody() != null &&
                    (
                        message.getBody().startsWith("subscribe ") ||
                        message.getBody().startsWith("http://") ||
                        message.getBody().startsWith("https://")
                    )
                    ) {

                String url = message.getBody().startsWith("subscribe ") ?
                        message.getBody().substring("subscribe ".length()).trim() :
                        message.getBody().trim();
                Log.debug("url = " + url);
                FeedSubscriptionRequest request = new FeedSubscriptionRequest(message.getFrom().toBareJID(), url);
                addFeed(request);

                Message reply = new Message();
                reply.setType(Message.Type.chat);
                reply.setTo(message.getFrom());
                reply.setFrom(bot);
                reply.setBody("I'm gonna do a quick check of that feed to make sure it exists and " +
                        "then I'll subscribe you. You'll hear from me shortly.");
                try {
                    componentManager.sendPacket(this, reply);
                }
                catch (Exception e) {
                    Log.error(e);
                }


            } else if (message.getBody() != null &&
                    message.getBody().startsWith("unsubscribe")) {

                String url = message.getBody().substring("unsubscribe".length()).trim();
                Message reply = new Message();
                reply.setType(Message.Type.chat);
                reply.setTo(message.getFrom());
                reply.setFrom(bot);
                if (deleteSubscription(url, message.getFrom().toBareJID())) {
                    reply.setBody("All set, you've been unsubscribed from '" + url + "'.");                        
                } else {
                    reply.setBody("It doesn't appear that you are subscribed to this feed. " +
                            "Type 'list' to view your feeds.");
                }

                try {
                    componentManager.sendPacket(this, reply);
                }
                catch (Exception e) {
                    Log.error(e);
                }
            } else if (message.getBody() != null &&
                    message.getBody().startsWith("off")) {

                boolean exists = updateHibernateState(message.getFrom().toBareJID(), true);
                if (!exists) {
                    Message error = new Message();
                    error.setType(Message.Type.chat);
                    error.setTo(message.getFrom());
                    error.setFrom(bot);
                    error.setBody("Whoops! I don't have you subscribed to any feeds. You'll" +
                        " have to subscribe to a feed before I turn off notifications.");
                    try {
                        componentManager.sendPacket(this, error);
                    }
                    catch (Exception e) {
                        Log.error(e);
                    }
                } else {
                    Message off = new Message();
                    off.setType(Message.Type.chat);
                    off.setTo(message.getFrom());
                    off.setFrom(bot);
                    off.setBody("All set. Let me know when you want me to turn you on again " +
                            "by replying with a message 'ON'.");
                    try {
                        componentManager.sendPacket(this, off);
                    }
                    catch (Exception e) {
                        Log.error(e);
                    }

                }


            } else if (message.getBody() != null &&
                    message.getBody().startsWith("on")) {

                boolean exists = updateHibernateState(message.getFrom().toBareJID(), false);

                if (!exists) {
                    Message error = new Message();
                    error.setType(Message.Type.chat);
                    error.setTo(message.getFrom());
                    error.setFrom(bot);
                    error.setBody("Whoops! I don't have you subscribed to any feeds. You'll" +
                        " have to subscribe to a feed before I turn on notifications.");
                    try {
                        componentManager.sendPacket(this, error);
                    }
                    catch (Exception e) {
                        Log.error(e);
                    }
                } else {
                    Message off = new Message();
                    off.setType(Message.Type.chat);
                    off.setTo(message.getFrom());
                    off.setFrom(bot);
                    off.setBody("All set. You should start receiving notifications soon!");
                    try {
                        componentManager.sendPacket(this, off);
                    }
                    catch (Exception e) {
                        Log.error(e);
                    }

                }

            } else if (message.getBody() != null &&
                    message.getBody().startsWith("status")) {

                FeedUser user = loadUser(message.getFrom().toBareJID());
                if (user == null) {
                    Message error = new Message();
                    error.setType(Message.Type.chat);
                    error.setTo(message.getFrom());
                    error.setFrom(bot);
                    error.setBody("Whoops! I don't have you subscribed to any feeds. You'll " +
                        "have to subscribe to at least one feed before I can tell you what " +
                        "your status is.");
                    try {
                        componentManager.sendPacket(this, error);
                    }
                    catch (Exception e) {
                        Log.error(e);
                    }
                } else {
                    Message off = new Message();
                    off.setType(Message.Type.chat);
                    off.setTo(message.getFrom());
                    off.setFrom(bot);
                    off.setBody(user.isHibernating() ? "You're currently in hibernation. Reply " +
                        "'on' to turn on notifications." :
                        "Your status is 'on'. Reply 'off' to turn off notifications.");
                    try {
                        componentManager.sendPacket(this, off);
                    }
                    catch (Exception e) {
                        Log.error(e);
                    }

                }


            } else if (message.getBody() != null &&
                    message.getBody().startsWith("help")) {
                Message help = new Message();
                help.setType(Message.Type.chat);
                help.setTo(message.getFrom());
                help.setFrom(bot);
                help.setBody("I provide RSS and Atom notifications via IM. If you subscribe " +
                        "to a feed, each time that site's feed is updated, " +
                        "you'll receive an IM from me with the changes.\n" +
                        "\t\n" +
                        "To communicate with me, you send short instant messages, I understand the following commands:\n" +
                        "\t\n" +
                        "\tsubscribe *url*\tSubscribes to the feed at *url*\n" +
                        "\tunsubscribe *url*\tUnsubscribes from the feed at *url*\n" +
                        "\tlist\tShows all the feeds you're subscribed to\n" +
                        "\t\n" +
                        "Some commands will require confirmation before they work. You'll get " +
                        "a URL to click on to verify that you want to perform the action.\n" +
                        "");
                try {
                    componentManager.sendPacket(this, help);
                }
                catch (Exception e) {
                    Log.error(e);
                }

            } else if (message.getBody() != null) {
                Message error = new Message();
                error.setType(Message.Type.chat);
                error.setTo(message.getFrom());
                error.setFrom(bot);
                error.setBody("Whoops! I didn't understand what you said. Try repeating it " +
                        "or say 'help' and I'll let you know what I do understand.");
                try {
                    componentManager.sendPacket(this, error);
                }
                catch (Exception e) {
                    Log.error(e);
                }
            }
        }
    }

    private void processPresence(Presence presence) {

        JID address = new JID(presence.getFrom().toBareJID());
        boolean allowed = true;
        if (!allowedUsers.isEmpty() && !allowedUsers.contains(address)) {
            allowed = false;
        }

        if (!allowedGroups.isEmpty()) {
            boolean foundUser = false;
            for (Group group : allowedGroups) {
                if (group.isUser(address)) {
                    foundUser = true;
                    break;
                }
            }
            if (!foundUser) {
                allowed = false;
            }
        }


        try {
            if (Presence.Type.subscribe == presence.getType()) {
                // Accept all presence requests if user has permissions
                // Reply that the subscription request was approved or rejected
                Presence reply = new Presence();
                reply.setTo(presence.getFrom());
                reply.setFrom(bot);
                reply.setType(allowed ? Presence.Type.subscribed : Presence.Type.unsubscribed);
                componentManager.sendPacket(this, reply);
            }
            else if (Presence.Type.unsubscribe == presence.getType()) {
                // Send confirmation of unsubscription
                Presence reply = new Presence();
                reply.setTo(presence.getFrom());
                reply.setFrom(bot);
                reply.setType(Presence.Type.unsubscribed);
                componentManager.sendPacket(this, reply);

                if (!allowed) {
                   // Send unavailable presence of the service
                   reply = new Presence();
                   reply.setTo(presence.getFrom());
                   reply.setFrom(presence.getTo());
                   reply.setType(Presence.Type.unavailable);
                   componentManager.sendPacket(this, reply);
               }
            }
            else if (Presence.Type.probe == presence.getType()) {
                // Send that the service is available
                Presence reply = new Presence();
                reply.setTo(presence.getFrom());
                reply.setFrom(bot);

                if (!allowed) {
                    // Send forbidden error since user is not allowed
                    reply.setError(PacketError.Condition.forbidden);
                }

                componentManager.sendPacket(this, reply);
            }
        }
        catch (ComponentException e) {
            Log.error(e);
        }
    }



    /**
     * Called by XML-RPC servlet when it is notified that a resource
     * has been modified.
     * @param feed the feed to fetch and update locally
     */
    public void addFeedUpdate(Feed feed) {
        existingFeeds.add(feed);
    }

    /**
     * Request by a user that a feed be added to the system.  Adds to a
     * queue of subscription requests.
     * @param request the feed and user information for the new subscription
     */
    public void addFeed(FeedSubscriptionRequest request) {
        newFeeds.add(request);
    }


    /**
     * Send notification to a user that a feed they subscribed to has been updated.
     * Checks to make sure that the user hasn't already been notified of the latest
     * entry in the feed.
     * <p>
     * If the user is in hibernate, the notification will not be sent.
     * 
     * @param user the user to send a notification to
     * @param feed the feed to send the notification about
     * @param force true to force the notification, false to let it fail if the user is hibernating
     *  or has already received the notification
     */
    public void sendNotification(FeedUser user, Feed feed, boolean force) {
        if (user.isHibernating()) {
            return;
        }

        
        synchronized ("notification-" + user.getJid().intern()) {
            // get the feed from the cache
            try {
                SyndFeed syndFeed = getFeed(feed.getURL(), true);
                // check the most recent entry pub date in the feed against the users most recent entry link for this feed
                long lastSentDate = new Date().getTime();
                //noinspection EmptyCatchBlock
                try {
                    lastSentDate = Long.parseLong(getSubscriptionLastEntry(user, feed));
                } catch (NumberFormatException e) {
                    // it's possible that the stored value is still a string value
                    // the string value was deprecated on 7/20/2007
                }
                List entries = syndFeed.getEntries();
                long mostRecentPubDate = ((SyndEntry)entries.get(0)).getPublishedDate().getTime();

                StringBuffer message = new StringBuffer();
                for (Object object : entries) {
                    SyndEntry entry = (SyndEntry) object;
                    if (entry.getPublishedDate().getTime() > lastSentDate) {
                        message.append(extractBody(entry));
                    }
                }

                // if we're forcing an update, then add the first entry
                if (force && message.length() == 0) {
                    message.append(extractBody(((SyndEntry)entries.get(0))));
                }


                if (force || message.length() > 0) {
                    Message reply = new Message();
                    reply.setType(Message.Type.chat);
                    reply.setTo(user.getJid());
                    reply.setFrom(bot);
                    reply.setBody("The feed '" + feed.getName() + "' has been updated: \n" +
                            message.toString());

                    try {
                        componentManager.sendPacket(this, reply);
                    }
                    catch (Exception e) {
                        Log.error(e);
                    }

                    // update the subscription to show that they've been notified of this feed update
                    updateSubscriptionLastEntry(user, feed, String.valueOf(mostRecentPubDate));
                }
            } catch (Exception e) {
                // since we're getting a cached version we should never get here...
            }
        }
    }

    /**
     * Send confirmation to a user that they have been subscribed to the given feed.
     * @param user the user to send a confirmation to
     * @param feed the feed to send a confirmation about
     */
    public void sendConfirmation(FeedUser user, Feed feed) {
        Message reply = new Message();
        reply.setType(Message.Type.chat);
        reply.setTo(user.getJid());
        reply.setFrom(bot);
        reply.setBody("All set.  You've been successfully subscribed to '" + feed.getName() + "'.");
        try {
            componentManager.sendPacket(this, reply);
        }
        catch (Exception e) {
            Log.error(e);
        }
    }


    /**
     * Send verification to a user asking for confirmation before subscribing user. Used
     * for subscriptions via XML-RPC / REST.  Subscriptions requested via IM are implicitly
     * valid.
     * @param subscriptionRequest the feed the user has requested a subscription to
     */
    public void sendVerification(FeedSubscriptionRequest subscriptionRequest) {
        String md5hash = StringUtils.hash(subscriptionRequest.getUrl() + "|" + subscriptionRequest.getJid());
        nonVerifiedRequests.putIfAbsent(md5hash, subscriptionRequest);

        Message reply = new Message();
        reply.setType(Message.Type.chat);
        reply.setTo(subscriptionRequest.getJid());
        reply.setFrom(bot);
        reply.setBody("Someone, possibly you, has submitted a request " +
                    "to subscribe to the following feed: \n\n" + subscriptionRequest.getUrl() + "\n\n" +
                    "To verify your subscription, click the following URL: " + "\n\n" +
                    serverName + "/http/subscriptions/verify?v=" + md5hash + "\n\n" +
                    "If you didn't sign up or did it on accident, just ignore this and you won't be contacted again.");
        try {
            componentManager.sendPacket(this, reply);
        }
        catch (Exception e) {
            Log.error(e);
        }
    }

    /**
     * Looks up the given verification code in the nonVerifiedRequests table. If a
     * FeedSubscriptionRequest exists for the given verification code, the method
     * return true and subscribes the user to the request feed. Otherwise the method
     * returns false.
     *
     * @param code the code to verify
     * @return feedsubscriptionrequest if the verification code exists in the system,
     *  null if the code doesn't exist.
     */
    public FeedSubscriptionRequest verifyRequest(String code) {
        if (nonVerifiedRequests.containsKey(code)) {
            FeedSubscriptionRequest fsr = nonVerifiedRequests.remove(code);
            Feed feed = loadFeed(fsr.getUrl());
            FeedUser user = loadUser(fsr.getJid());

            Connection con = null;
            PreparedStatement pstmt = null;
            try {
                con = DbConnectionManager.getConnection();
                pstmt = con.prepareStatement(UPDATE_SUBSCRIPTION_VERIFIED);
                pstmt.setLong(1, user.getID());
                pstmt.setLong(2, feed.getID());
                pstmt.executeUpdate();
            }
            catch (Exception sqle) {
                // ignore
            }
            finally {
                DbConnectionManager.closeConnection(pstmt, con);
            }

            sendConfirmation(user, feed);

            return fsr;
        } else {
            return null;
        }
    }



    /**
     * Send error message to user: feed requested is not valid.
     * @param user user to send the message to
     * @param message the message to send the user
     */
    public void sendFeedErrorNotification(FeedUser user, String message) {
        Message reply = new Message();
        reply.setType(Message.Type.chat);
        reply.setTo(user.getJid());
        reply.setFrom(bot);
        reply.setBody(message);
        try {
            componentManager.sendPacket(this, reply);
        }
        catch (Exception e) {
            Log.error(e);
        }
    }


    /**
     * Gets the number of users who are subscribed to the given feed / url.
     * @param feedURL the url to check for subscribers
     * @return the number of users subscribed to the given endpoint
     */
    public int getUsersCount(String feedURL) {
        int count = 0;
        Connection con = null;
        PreparedStatement pstmt = null;
        try {
            con = DbConnectionManager.getConnection();
            pstmt = con.prepareStatement(GET_USERS_COUNT_BY_FEED);
            pstmt.setString(1, feedURL);
            ResultSet rs = pstmt.executeQuery();
            while (rs.next()) {
                count = rs.getInt(1);
            }
        }
        catch (Exception sqle) {
            Log.debug(sqle);
        }
        finally {
            DbConnectionManager.closeConnection(pstmt, con);
        }

        return count;
    }


    /**
     * Get all the users subscribed to a particular feed.
     * @param feed the feed to find subscriptions for
     * @return a list of users who have subscribed to this feed
     */
    public List<FeedUser> getUsers(Feed feed) {
        List<FeedUser> users = new ArrayList<FeedUser>();
        Connection con = null;
        PreparedStatement pstmt = null;
        try {
            con = DbConnectionManager.getConnection();
            pstmt = con.prepareStatement(GET_USERS_BY_FEED);
            pstmt.setLong(1, feed.getID());
            ResultSet rs = pstmt.executeQuery();
            while (rs.next()) {
                long userID = rs.getLong(1);
                String jid = rs.getString(2);
                int hibernating = rs.getInt(3);
                FeedUser user = new FeedUser(userID, jid, hibernating > 0);
                users.add(user);
            }
        }
        catch (Exception sqle) {
            Log.debug(sqle);
        }
        finally {
            DbConnectionManager.closeConnection(pstmt, con);
        }

        return users;
    }

    /**
     * Create a user in the system.
     * @param jid the JID of the user to create
     * @return a new user
     */
    private FeedUser createUser(String jid) {
        long userID = SequenceManager.nextID(JIVEIDFEEDUSER);

        Connection con = null;
        PreparedStatement pstmt = null;
        try {
            con = DbConnectionManager.getConnection();
            pstmt = con.prepareStatement(CREATE_USER);
            pstmt.setLong(1, userID);
            pstmt.setString(2, jid);

            pstmt.executeUpdate();
        }
        catch (Exception sqle) {
            Log.debug(sqle);
        }
        finally {
            DbConnectionManager.closeConnection(pstmt, con);
        }

        return new FeedUser(userID, jid, false);
    }

    private FeedUser loadUser(String jid) {
        FeedUser user = null;
        Connection con = null;
        PreparedStatement pstmt = null;
        try {
            con = DbConnectionManager.getConnection();
            pstmt = con.prepareStatement(GET_USER_BY_JID);
            pstmt.setString(1, jid);
            ResultSet rs = pstmt.executeQuery();
            while (rs.next()) {
                long userID = rs.getLong(1);
                int hibernating = rs.getInt(3);
                user = new FeedUser(userID, jid, hibernating > 0);

            }
        }
        catch (Exception sqle) {
            Log.debug(sqle);
        }
        finally {
            DbConnectionManager.closeConnection(pstmt, con);
        }
        return user;
    }



    private Feed createFeed(String url, String name, String mostRecentEntry) {
        long feedID = SequenceManager.nextID(JIVEIDFEED);
        Connection con = null;
        PreparedStatement pstmt = null;
        try {
            con = DbConnectionManager.getConnection();
            pstmt = con.prepareStatement(CREATE_FEED);
            pstmt.setLong(1, feedID);
            pstmt.setString(2, name);
            pstmt.setString(3, url);
            pstmt.setString(4, mostRecentEntry);
            pstmt.executeUpdate();
        }
        catch (Exception sqle) {
            Log.debug(sqle);
        }
        finally {
            DbConnectionManager.closeConnection(pstmt, con);
        }

        return new Feed(feedID, url, name, mostRecentEntry);
    }

    public Feed loadFeed(String url) {
        Feed feed =  null;
        Connection con = null;
        PreparedStatement pstmt = null;
        try {
            con = DbConnectionManager.getConnection();
            pstmt = con.prepareStatement(GET_FEED_BY_URL);
            pstmt.setString(1, url);
            ResultSet rs = pstmt.executeQuery();
            while (rs.next()) {
                long feedID = rs.getLong(1);
                String feedName = rs.getString(2);
                String feedURL = rs.getString(3);
                String mostRecentEntry = rs.getString(4);
                feed = new Feed(feedID, feedName, feedURL, mostRecentEntry);
            }
        }
        catch (Exception sqle) {
            Log.debug(sqle);
        }
        finally {
            DbConnectionManager.closeConnection(pstmt, con);
        }
        return feed;
    }

    /**
     * Get all the feeds.
     * @return a list of all the feeds in the system.
     */
    public List<Feed> loadFeeds() {
        List<Feed> feeds = new ArrayList<Feed>();
        Connection con = null;
        PreparedStatement pstmt = null;
        try {
            con = DbConnectionManager.getConnection();
            pstmt = con.prepareStatement(GET_FEEDS);
            ResultSet rs = pstmt.executeQuery();
            while (rs.next()) {
                long feedID = rs.getLong(1);
                String name = rs.getString(2);
                String url = rs.getString(3);
                String mostRecentEntry = rs.getString(4);
                Feed feed = new Feed(feedID, name, url, mostRecentEntry);
                feeds.add(feed);
            }
        }
        catch (Exception sqle) {
            Log.debug(sqle);
        }
        finally {
            DbConnectionManager.closeConnection(pstmt, con);
        }

        return feeds;
    }


    /**
     * Get all the feeds a user has subscribed to.
     * @param user the user to retrieve subscriptions for
     * @return a list of feeds the user has subscribed to
     */
    private List<Feed> getFeeds(FeedUser user) {
        List<Feed> feeds = new ArrayList<Feed>();
        Connection con = null;
        PreparedStatement pstmt = null;
        try {
            con = DbConnectionManager.getConnection();
            pstmt = con.prepareStatement(GET_FEEDS_BY_USER);
            pstmt.setLong(1, user.getID());
            ResultSet rs = pstmt.executeQuery();
            while (rs.next()) {
                long feedID = rs.getLong(1);
                String name = rs.getString(2);
                String url = rs.getString(3);
                String mostRecentEntry = rs.getString(4);
                Feed feed = new Feed(feedID, name, url, mostRecentEntry);
                feeds.add(feed);
            }
        }
        catch (Exception sqle) {
            Log.debug(sqle);
        }
        finally {
            DbConnectionManager.closeConnection(pstmt, con);
        }

        return feeds;
    }

    private void updateFeed(Feed feed) {
        Connection con = null;
        PreparedStatement pstmt = null;
        try {
            con = DbConnectionManager.getConnection();
            pstmt = con.prepareStatement(UPDATE_FEED);
            pstmt.setString(1, feed.getName());
            pstmt.setString(2, feed.getURL());
            pstmt.setString(3, feed.getMostRecentEntry());
            pstmt.setLong(4, feed.getID());
            pstmt.executeUpdate();
        }
        catch (Exception sqle) {
            // ignore
        }
        finally {
            DbConnectionManager.closeConnection(pstmt, con);
        }
    }

    /**
     * Sets a user's hibernate state to on or off.
     * @param jid the jid of the user to update the state for
     * @param hibernate true to turn hibernate on, false otherwise
     * @return true if the user exists, false otherwise
     */
    private boolean updateHibernateState(String jid, boolean hibernate) {

        FeedUser user = loadUser(jid);
        if (user != null) {
            Connection con = null;
            PreparedStatement pstmt = null;
            try {
                con = DbConnectionManager.getConnection();
                pstmt = con.prepareStatement(UPDATE_USER_SLEEP_STATE);
                pstmt.setInt(1, hibernate ? 1 : 0);
                pstmt.setLong(2, user.getID());
                pstmt.executeUpdate();
            }
            catch (Exception sqle) {
                // ignore
            }
            finally {
                DbConnectionManager.closeConnection(pstmt, con);
            }
            return true;
        } else {
            return false;
        }
    }

    /**
     * Attempts to autodiscover feeds in a given URL.
     * @param feedURL the url to discover feeds in
     * @return a list of feed URL's available from the given URL, an
     * empty list if no feeds were found.
     *
     */
    public List<String> discoverFeeds(String feedURL) {
        List<String> links = new ArrayList<String>();


        try {
            Parser parser = new Parser (feedURL);
            NodeList list = new NodeList ();
            NodeFilter filter = new AndFilter(
                new TagNameFilter ("LINK"),
                    new HasAttributeFilter("REL"));
            for (NodeIterator e = parser.elements (); e.hasMoreNodes (); )
                e.nextNode ().collectInto (list, filter);

            for (SimpleNodeIterator nodes = list.elements(); nodes.hasMoreNodes(); ) {
                Tag tag = (Tag)nodes.nextNode();
                if (tag.getAttribute("rel") != null &&
                        tag.getAttribute("rel").equalsIgnoreCase("alternate") &&
                        tag.getAttribute("href") != null) {
                    String feed = tag.getAttribute("href");
                    links.add(feed);
                }
                
            }
        } catch (ParserException e) {
            /* ignore */
        }


        return links;
    }


    private SyndFeed getFeed(String feedURL, boolean useCache)
            throws MalformedURLException, IllegalArgumentException, FeedException,
            FetcherException {

        FeedFetcherCache feedInfoCache = HashMapFeedInfoCache.getInstance();
        SyndFeed feed;
        URL url = null;
        if (proxyHost.length() > 0 && proxyPort > 0) {
            url = new URL("http", proxyHost, proxyPort, feedURL);
        } else {
            url = new URL(feedURL);
        }

        

        if (useCache && feedInfoCache.getFeedInfo(url) != null) {
            return feedInfoCache.getFeedInfo(url).getSyndFeed();
        } else {
            FeedFetcher feedFetcher = new InstantFeedsFeedFetcher(feedInfoCache);
            int n = getUsersCount(feedURL);
            feedFetcher.setUserAgent("instantFeeds " + 
                    pluginManager.getVersion(this) +
                    "; (+http://cephas.net/projects/instantfeeds/; " +
                    n + " subscribers)");

            try {
                feed = feedFetcher.retrieveFeed(url);
            } catch (IOException e) {
                throw new FetcherException(e.getCause());
            }
            return feed;
        }
    }



    

    private void createSubscription(FeedUser user, Feed feed, boolean verification, String mostRecentEntry) {
        // TODO: before creating, check to see if the user / feed combo exists...

        Connection con = null;
        PreparedStatement pstmt = null;
        try {
            con = DbConnectionManager.getConnection();
            pstmt = con.prepareStatement(CREATE_SUBSCRIPTION);
            pstmt.setLong(1, user.getID());
            pstmt.setLong(2, feed.getID());
            pstmt.setInt(3, verification ? 1 : 0);
            pstmt.setString(4, mostRecentEntry);
            pstmt.executeUpdate();
        }
        catch (Exception sqle) {
            Log.debug(sqle);
        }
        finally {
            DbConnectionManager.closeConnection(pstmt, con);
        }
    }

    private String getSubscriptionLastEntry(FeedUser user, Feed feed) {
        String mostRecentEntry = "";
        Connection con = null;
        PreparedStatement pstmt = null;
        try {
            con = DbConnectionManager.getConnection();
            pstmt = con.prepareStatement(GET_SUBSCRIPTION_LAST_ENTRY);
            pstmt.setLong(1, user.getID());
            pstmt.setLong(2, feed.getID());
            ResultSet rs = pstmt.executeQuery();
            while (rs.next()) {
                mostRecentEntry = rs.getString(1);
            }
        }
        catch (Exception sqle) {
            Log.debug(sqle);
        }
        finally {
            DbConnectionManager.closeConnection(pstmt, con);
        }
        return mostRecentEntry;
    }

    /**
     * Update a feed subscription for the given user to reflect the most recently created
     * entry.
     * @param user the user associated with this subscription
     * @param feed the feed the user is subscribed to
     * @param mostRecentEntry the most recent link sent to the user
     */
    private void updateSubscriptionLastEntry(FeedUser user, Feed feed, String mostRecentEntry) {
        Connection con = null;
        PreparedStatement pstmt = null;
        try {
            con = DbConnectionManager.getConnection();
            pstmt = con.prepareStatement(UPDATE_SUBSCRIPTION_LAST_ENTRY);
            pstmt.setString(1, mostRecentEntry);
            pstmt.setLong(2, feed.getID());
            pstmt.setLong(3, user.getID());
            pstmt.executeUpdate();
        }
        catch (Exception sqle) {
            // ignore
        }
        finally {
            DbConnectionManager.closeConnection(pstmt, con);
        }

    }

    private boolean deleteSubscription(String url, String jid) {

        Feed feed;
        FeedUser user;
        try {
            feed = loadFeed(url);
            user = loadUser(jid);
        } catch (IllegalArgumentException iae) {
            return false;
        }

        Connection con = null;
        PreparedStatement pstmt = null;
        try {
            con = DbConnectionManager.getConnection();
            pstmt = con.prepareStatement(DELETE_SUBSCRIPTION);
            pstmt.setLong(1, user.getID());
            pstmt.setLong(2, feed.getID());
            pstmt.executeUpdate();
        }
        catch (Exception sqle) {
            Log.debug("error deleting feed:" + sqle.toString());
        }
        finally {
            DbConnectionManager.closeConnection(pstmt, con);
        }

        return true;
    }

    /**
     * Turns a comma-delimitted list of strings into a Collection of Users/JIDs.
     *
     * @param str the String.
     * @return a list.
     */
    private List<JID> userStringToList(String str) {
        List<JID> values = new ArrayList<JID>();
        StringTokenizer tokens = new StringTokenizer(str, ",");
        while (tokens.hasMoreTokens()) {
            String value = tokens.nextToken().trim();
            if (!value.equals("")) {
                // See if this is a full JID or just a username.
                if (value.contains("@")) {
                    values.add(new JID(value));
                }
                else {
                    values.add(XMPPServer.getInstance().createJID(value, null));
                }
            }
        }
        return values;
    }

    /**
     * Turns a comma-delimitted list of strings into a Collection of Groups.
     *
     * @param str the String.
     * @return a list.
     */
    private List<Group> groupStringToList(String str) {
        List<Group> values = new ArrayList<Group>();
        StringTokenizer tokens = new StringTokenizer(str, ",");
        while (tokens.hasMoreTokens()) {
            String value = tokens.nextToken().trim();
            if (!value.equals("")) {
                try {
                    Group group = groupManager.getGroup(value);
                    values.add(group);
                } catch (GroupNotFoundException e) {
                    Log.error("error adding group '" + value + "' to instantfeeds: " + e.toString());
                }
            }
        }
        return values;
    }


    /**
     * Extract the title, link and body of the feed entry.
     * @param entry the SyndEntry to extract the body from
     * @return entry a fragment that contains the title + linebreak + link + linebreak +
     * first 255 chars of the body stripped of any tags.
     */
    private String extractBody(SyndEntry entry) {
        String link = entry.getLink();
        String title = entry.getTitle();
        SyndContent contentBody = entry.getDescription();
        String body;

        body = ParserUtils.trimAllTags(contentBody.getValue(), false);
        body = StringEscapeUtils.unescapeHtml(body);

        if (body.length() > 255) {
            int lastSpace = body.substring(0, 255).lastIndexOf(" ");
            if (lastSpace > -1) {
                body = body.substring(0, lastSpace) + "...";
            } else {
                body = body.substring(0, 255) + "...";
            }
        }

        return title + "\n" + link + "\n" + body + "\n\n";
    }



    /**
     * Polls the queue of new feed subscription requests, attempts to load
     * the requested feed and then either sends a confirmation to the user
     * and subscribes them to the feed (if the feed is valid) or sends
     * an error message if the feed is invalid.
     */
    private class NewFeedsTask implements Runnable {
        private int id;
        private boolean running;

        NewFeedsTask(int id) {
            this.id = id;
            Log.debug("#NewFeedsTask:" + this.id + " initializing");
            running = true;
        }

        public void run() {
            while(running) {
                if (shouldRun()) {
                    while(!newFeeds.isEmpty()) {
                        FeedSubscriptionRequest f = newFeeds.poll();

                        if (f != null) {

                            // see if feed exists already in db
                            Feed feed = null;
                            FeedUser user = null;
                            try {
                                feed = loadFeed(f.getUrl());
                            } catch (Exception iae) {
                                // ignore
                                Log.debug("#NewsFeedsTask[" + id + "]: error loading feed: " + f.getUrl());
                            }

                            try {
                                user = loadUser(f.getJid());
                            } catch (Exception iae) {
                                // ignore
                                Log.debug("#NewsFeedsTask[" + id + "]: error loading user: " + f.getJid());
                            }

                            Log.debug("#NewsFeedsTask[" + id + "]: feedsub request[" + f.getJid() +"," + f.getUrl() + "]");


                            // feed doesn't exist in DB
                            if (feed == null) {

                                boolean isFeed = false;
                                boolean containsAutodiscoverFeeds = false;
                                List<String> availableFeeds = null;
                                boolean isProtected = false;
                                String feedName = null;
                                SyndFeed syndFeed = null;

                                try {
                                    syndFeed = getFeed(f.getUrl(), false);
                                } catch (MalformedURLException e) {
                                    Log.debug("invalid URL: " + e.toString());
                                } catch (FeedException e) {
                                    Log.debug("feedException: " + e.toString());
                                    // invalid feed.. might be a HTML page
                                    // see if there are any feeds embedded in the page
                                    availableFeeds = discoverFeeds(f.getUrl());
                                    if (availableFeeds != null && availableFeeds.size() > 0) {
                                        containsAutodiscoverFeeds = true;
                                    }



                                } catch (FetcherException e) {
                                    // check for 401 status code
                                    if (e.getResponseCode() == 401) {
                                        isProtected = true;
                                    } else {
                                        Log.debug("fetcherException: " + e.toString());
                                    }
                                }


                                if (syndFeed != null) {
                                    feedName = syndFeed.getTitle();
                                    isFeed = true;
                                }

                                // exists?
                                if (isFeed) {

                                    // create user?
                                    if (user == null) {
                                        user = createUser(f.getJid());
                                    }

                                    // create feed
                                    SyndEntry firstEntry = (SyndEntry)syndFeed.getEntries().get(0);
                                    feed = createFeed(f.getUrl(), feedName, firstEntry.getLink());

                                    // associate user with feed
                                    SyndEntry mostRecentEntry = (SyndEntry)syndFeed.getEntries().get(0);
                                    createSubscription(user, feed, f.isVerified(), mostRecentEntry.getLink());

                                    // send verification?
                                    if (!f.isVerified()) {
                                        sendVerification(f);

                                    } else {
                                        // send confirmation
                                        sendConfirmation(user, feed);
                                    }

                                // feed fetch failed..
                                } else {

                                    // create user?
                                    if (user == null) {
                                        user = createUser(f.getJid());
                                    }


                                    // send message to user: not a feed
                                    if (containsAutodiscoverFeeds && availableFeeds != null) {
                                        StringBuffer message = new StringBuffer();
                                        message.append("There are multiple feeds available at the " +
                                            "URL you provided. Please choose one of the following feeds: \n\n");
                                        for (String availableFeed : availableFeeds) {
                                            message.append(availableFeed).append("\n");
                                        }
                                        sendFeedErrorNotification(user, message.toString());



                                    } else if (isProtected) {
                                        String message = "The feed you requested: '" +
                                                f.getUrl() + "' requires login credentials. You " +
                                                "can use the format http://user:pass@" + f.getUrl()
                                                .replaceAll("http://", "")
                                                .replaceAll("https://", "") +
                                                " to request a protected feed.";
                                        sendFeedErrorNotification(user, message);

                                    } else {
                                        String message = "The feed you requested: '" +
                                                f.getUrl() + "' does not appear to be a valid feed.";
                                        sendFeedErrorNotification(user, message);

                                    }


                                }
                            } else {
                            // feed exists in DB
                                try {
                                    SyndFeed syndFeed = getFeed(f.getUrl(), true);

                                    // has the user subscribed?
                                    boolean exists = false;

                                    // create user?
                                    if (user == null) {
                                        user = createUser(f.getJid());
                                    // user exists, check to make sure they haven't already subscribed
                                    } else {
                                        // TODO: find out if user has already subscribed

                                        List<Feed> usersFeeds = getFeeds(user);
                                        for (Feed userFeed : usersFeeds) {
                                            if (userFeed.getURL().equalsIgnoreCase(f.getUrl())) {
                                                exists = true;
                                                break;
                                            }
                                        }
                                    }

                                    // user has subscribed already...
                                    if (exists) {
                                        String message = "You are already subscribed to the feed: \n\n" +
                                                f.getUrl() + "\n\n Please try again.";
                                        sendFeedErrorNotification(user, message);


                                    // new subscribtion...
                                    } else {
                                        // associate user with feed
                                        SyndEntry mostRecentEntry = (SyndEntry)syndFeed.getEntries().get(0);
                                        createSubscription(user, feed, f.isVerified(), mostRecentEntry.getLink());

                                        // send verification?
                                        if (!f.isVerified()) {
                                            sendVerification(f);

                                        } else {
                                            // send confirmation
                                            sendConfirmation(user, feed);
                                        }
                                    }


                                } catch (Exception e) {
                                    // retrieving the item from cache... should never hit this
                                }
                            }

                        }
                    }
                }

                // Sleep for a few seconds before we loop through and see if we
                // need to run again
                try {
                    Thread.sleep(5 * JiveConstants.SECOND);
                }
                catch (InterruptedException e) {
                    Log.debug(e.getMessage(), e);
                }


            }
        }

        public boolean shouldRun() {
            return newFeeds.size() > 0;
        }

        public synchronized void stop() {
            this.running = false;
        }

    }

    /**
     * Loads all existing feeds from the database into the existingFeeds
     * queue for the pollers and then sleeps for n minutes.
     */
    private class LoadFeedsTask implements Runnable {
        private boolean running;
        private long pollingInterval;

        LoadFeedsTask(long pollInt) {
            pollingInterval = pollInt;
            Log.debug("#LoadFeedsTask initializing with pollInterval of " + pollInt);
            running = true;
        }

        public void run() {
            while (running) {
                List<Feed> feeds = loadFeeds();
                for (Feed feed : feeds) {
                    if (!existingFeeds.contains(feed)) {
                        existingFeeds.add(feed);
                    }
                }
                
                try {
                    Thread.sleep(pollingInterval);
                }
                catch (InterruptedException e) {
                    Log.debug(e.getMessage(), e);
                }

            }

        }

        public synchronized void stop() {
            this.running = false;
        }

    }

    private class PollFeedsTask implements Runnable {
        private int id;
        private boolean running;

        PollFeedsTask(int id) {
            this.id = id;
            Log.debug("#SendNotificationsTask:" + this.id + " initializing");
            running = true;

        }

        public void run() {
            while(running) {
                if (shouldRun()) {
                    while(!existingFeeds.isEmpty()) {
                        Feed f = existingFeeds.poll();
                        try {
                            SyndFeed sf = getFeed(f.getURL(), false);
                            if (sf.getEntries().size() > 0) {
                                SyndEntry firstEntry = (SyndEntry)sf.getEntries().get(0);
                                if (!firstEntry.getLink().equalsIgnoreCase(f.getMostRecentEntry())) {
                                    f.setMostRecentEntry(firstEntry.getLink());
                                    // save the most recent link on the feed
                                    updateFeed(f);
                                    updatedFeeds.add(f);
                                }
                            }
                        } catch (Exception e) {
                            // existing feeds... if we have an exception.. let it go for now
                            // TODO: keep track of how many errors we get for a feed and remove it
                            // if the number exceeds n
                        }
                    }
                }

                try {
                    Thread.sleep(5 * JiveConstants.SECOND);
                }
                catch (InterruptedException e) {
                    Log.debug(e.getMessage() + " on thread[" + id + "]", e);
                }
            }

        }

        public boolean shouldRun() {
            return existingFeeds.size() > 0;

        }

        public synchronized void stop() {
            this.running = false;
        }

    }

    /**
     * Polls the queue of updated feeds, retrieves users who are subscribed to
     * the feed and sends a notification to each one. Sleeps for 5 seconds
     * between each check of the queue.
     */
    private class SendNotificationsTask implements Runnable {
        private int id;
        private boolean running;

        SendNotificationsTask(int id) {
            this.id = id;
            Log.debug("#SendNotificationsTask:" + this.id + " initializing");
            running = true;
        }

        public void run() {
            while(running) {
                if (shouldRun()) {
                    while (!updatedFeeds.isEmpty()) {
                        Feed f = updatedFeeds.poll();
                        List<FeedUser> users = getUsers(f);
                        for (FeedUser user : users) {
                            sendNotification(user, f, false);
                        }
                    }
                }

                try {
                    Thread.sleep(5 * JiveConstants.SECOND);
                }
                catch (InterruptedException e) {
                    Log.debug("#SendNotificationsTask: " + this.id + ":" + e.getMessage(), e);
                }
                
            }
        }

        public boolean shouldRun() {
            return updatedFeeds.size() > 0;

        }

        public synchronized void stop() {
            this.running = false;
        }
    }

    private class FakeConnection extends VirtualConnection {
        FakeConnection() {
        }



        public void closeVirtualConnection() {
            // do nothing
        }

        public InetAddress getInetAddress() throws UnknownHostException {
            return InetAddress.getLocalHost();
        }

        public void systemShutdown() {
            // ignore
        }

        public void deliver(Packet packet) throws UnauthorizedException {
            // ignore
        }

        public void deliverRawText(String text) {
            // ignore
        }

        public byte[] getAddress() throws UnknownHostException {
            return InetAddress.getLocalHost().getAddress();
        }


        public String getHostAddress() throws UnknownHostException {
            return InetAddress.getLocalHost().getHostAddress();
        }

        public String getHostName() throws UnknownHostException {
            return InetAddress.getLocalHost().getHostName();
        }
    }

}
